
#include <assert.h>
#include "line_factory.hh"

namespace network_config
{
  typedef std::string::size_type str_index;

  /* skeleton

  Subsection*  SubsectionInstance::GetSubsection (kvtree::TreeManager* manager)
  {
    if (!EnsureInterfaceSubsection (manager, kvtree::node_types::) )
      manager -> OpenNewSubsection (new Subsection (manager -> getCurrent ()));
    
    kvtree::Group* ret = manager -> getCurrent ();
    assert (ret -> Type () == kvtree::node_types::);
    return (Subsection*) ret;
  }

  bool ::Match (std::string& content, kvtree::TreeManager* manager)
  {
    Subsection* current;

    if (tokens[0]=="") {
      current = GetSubsection (manager);

      return true;

    } else if (tokens[0]=="") {
      current = GetSubsection (manager);

      return true;
 
    }

    return false;
  }


  */

  std::vector <std::string> tokens;

  void DeQuote (std::string& str); // defined in network_config

  void Tokenize (const std::string& line)
  { // TODO: handle different quote types and do some error handling
    // for unterminated quotes and eol escapes
    tokens.clear ();

    str_index token_start=0;
    str_index token_end=0;

    bool in_quotes = false;
    bool concatenate_next = false;

    do {
      token_start = line.find_first_not_of("\t ", token_end);
      if (token_start != std::string::npos) {
	std::string token;

	token_end = line.find_first_of("\t ", token_start);
	if (token_end == std::string::npos)
	  token = line.substr(token_start);
	else
	  token = line.substr(token_start, token_end - token_start);

	if (!concatenate_next)
	  tokens.push_back (token);
	else { // if we are in a quoted part, append separator + token to successor
	  tokens.back() += (line[token_start - 1] == '\t') ? "\t" : " ";
	  tokens.back() += token;
	} 
	
	// count unescaped quotes to see if we have to toggle quote mode.
	int real_quotes=0;
	str_index quote=0;
	do {
	  quote=token.find_first_of("\"'", quote);
	  if (quote != std::string::npos) {
	    if (quote==0 || token[quote - 1] != '\\')
	      real_quotes++;
	    quote++;
	  }
	} while (quote != std::string::npos && quote != token.length());
	
	if (real_quotes % 2 == 1)
	  in_quotes = !in_quotes;
	concatenate_next = in_quotes || (token[token.size() - 1] == '\\');

      } else
	token_end = std::string::npos;

    } while (token_end != std::string::npos);

    // now we translate all '-' into '_' for the first token
    str_index minus;
    while ((minus = tokens[0].find_first_of ("-")) != std::string::npos)
      tokens[0][minus] = '_';

    // de-quote remaining tokens
    for (unsigned int i = 1; i < tokens.size (); i++)
      DeQuote (tokens[i]);
      
  }



  bool DeComment::Match (std::string& content, kvtree::TreeManager* manager)
  {
    str_index content_end;
    //unsigned int content_start = 0;
    str_index lenght = content.length();

    //for ( ; 
    //  content_start < lenght &&
    //	    (content[content_start]==' ' || content[content_start]=='\t') ;
    //	  content_start++ );

    str_index content_start = content.find_first_not_of ("\t ");

    for ( content_end = content_start; 
	  content_end < lenght &&
	    !(content[content_end] =='#' && (content_end == 0 || content[content_end - 1]=='\t')) ;
	  content_end++ );

    if (content_start == content_end)
      return true; // there is no relevant content

    // copy relevant content
    content = content.substr (content_start, content_start - content_end);
    Tokenize (content);
    return false;
  }

  bool InterfaceSectionInstance::Match (std::string& content, kvtree::TreeManager* manager)
  { // TODO: multiple interfaces !
    if (tokens[0] == "interface") {
      std::string profile="default";
      std::string iface="eth0";

      if (tokens.size () == 2) {
	const std::string& arg=tokens[1];

	str_index profile_start = arg.find_first_of("(");
	if (profile_start != std::string::npos) {
	  str_index profile_end = arg.find_first_of(")");
	  if (profile_end != arg.length() -1) {
	    manager -> ParserWarning ("malformed profile specifier.");
	    if (profile_end == std::string::npos)
	      profile_end == arg.length()-1;
	  }

	  iface = arg.substr(0, profile_start);
	  profile = arg.substr (profile_start + 1, profile_end - profile_start -1);
	  if (profile == "") {
	    manager -> ParserWarning ("empty profile specifier, assuming 'default'.");
	    profile="default";
	  }
	    
	} else
	  iface = arg;
      }
      else if (tokens.size() < 2)
	manager -> ParserWarning ("interface section requires interface id. assuming 'eth0'.");
      else
	manager -> ParserWarning ("ignoring trailing arguments.");

      kvtree::Group* root = manager -> getRoot ();
      manager -> MoveCurrent (root);
      manager -> OpenNewSubsection (new InterfaceSection (iface, profile, root));

      return true;
    }
    return false;
  }


  
  bool EnsureInterfaceSubsection (kvtree::TreeManager* manager, kvtree::node_type wanted)
  { 
    if (! manager -> SwitchToParallelSection (kvtree::node_types::interface_section, wanted) ) {
      if (manager -> getCurrent () -> Type () != kvtree::node_types::interface_section) {
	assert (manager -> getCurrent () -> Type () == kvtree::node_types::global);
	manager -> ParserWarning ("interface expected, setting to eth0(orphaned)");
	manager -> OpenNewSubsection (new InterfaceSection ("eth0", "orphaned", manager -> getCurrent ()));
      }
      assert (manager -> getCurrent () -> Type () == kvtree::node_types::interface_section);
      return false;
    }

    assert (manager -> getCurrent () -> Type () == wanted);
    return true;
  }


  AddressSubsection*  AddressSubsectionInstance::GetAddressSubsection (kvtree::TreeManager* manager)
  {
    if (!EnsureInterfaceSubsection (manager, kvtree::node_types::address_subsection) )
      manager -> OpenNewSubsection (new AddressSubsection (manager -> getCurrent ()));

    kvtree::Group* ret = manager -> getCurrent ();
    assert (ret -> Type () == kvtree::node_types::address_subsection);
    return (AddressSubsection*) ret;
  }

  bool AddressSubsectionInstance::Match (std::string& content, kvtree::TreeManager* manager)
  {
    AddressSubsection* current;
  
    if (tokens[0] == "ip") {
      current=GetAddressSubsection (manager);

      IPLine* ip_line = new IPLine (content);
      if (tokens.size () < 2 || ! (ip_line -> address.FromString (tokens[1])) ) {
	manager -> ParserWarning ("missing or malformed CIDR");
	ip_line -> Set ();
      } else if (tokens.size () > 2)
	manager -> ParserWarning ("ignoring trailing arguments.");

      current -> childs.push_back (ip_line);

      return true;

    } else if (tokens [0] == "gw") {
      current=GetAddressSubsection (manager);

      GatewayLine* gw_line = new GatewayLine (content);
      if (tokens.size () < 2 || ! (gw_line -> nexthop.FromString (tokens[1])) ) {
	manager -> ParserWarning ("missing or malformed ip address for nexthop");
	gw_line -> Set ();
      } else if (tokens.size () > 2)
	manager -> ParserWarning ("ignoring trailing arguments.");

      current -> childs.push_back (gw_line);

      return true;

    } else if (tokens [0] == "nameserver") {
      current=GetAddressSubsection (manager);

      NameserverLine* ns_line = new NameserverLine (content);
      if (tokens.size () < 2 || ! (ns_line -> nameserver.FromString (tokens[1])) ) {
	manager -> ParserWarning ("missing or malformed ip address for nameserver");
	ns_line -> Set ();
      } else if (tokens.size () > 2)
	manager -> ParserWarning ("ignoring trailing arguments.");

      current -> childs.push_back (ns_line);

      return true;

    } else if (tokens [0] == "search") {
      current=GetAddressSubsection (manager);

      if (tokens.size () < 2) {
	manager -> ParserWarning ("missing domain");
      } else {
	if (tokens.size () > 2)
	  manager -> ParserWarning ("ignoring trailing arguments.");

	SearchLine* s_line = new SearchLine (content);
	s_line -> domain = tokens [1];
	current -> childs.push_back (s_line);
      }

      return true;

    } else if (tokens [0] == "dhcp") {
      current=GetAddressSubsection (manager);
      current -> setDhcp (true);
      return true;

    }

    return false;
  }


  WirelessSubsection*  WirelessSubsectionInstance::GetWirelessSubsection (kvtree::TreeManager* manager)
  {
    if (!EnsureInterfaceSubsection (manager, kvtree::node_types::wireless_subsection) )
      manager -> OpenNewSubsection (new WirelessSubsection (manager -> getCurrent ()));

    kvtree::Group* ret = manager -> getCurrent ();
    assert (ret -> Type () == kvtree::node_types::wireless_subsection);
    return (WirelessSubsection*) ret;
  }


  bool WirelessSubsectionInstance::Match (std::string& content, kvtree::TreeManager* manager)
  {
    WirelessSubsection* current;

    if (tokens[0]=="mode") {
      current = GetWirelessSubsection (manager);

      ModeLine* m_line = new ModeLine (content);
      if (tokens.size () < 2 || ! (m_line -> FromString (tokens[1])) ) {
	manager -> ParserWarning ("missing or malformed operation mode");
	m_line -> Set ();
      } else if (tokens.size () > 2)
	manager -> ParserWarning ("ignoring trailing arguments.");

      current -> childs.push_back (m_line);

      return true;

    } else if (tokens[0]=="channel") {
      current = GetWirelessSubsection (manager);

      ChannelLine* c_line = new ChannelLine (content);
      if (tokens.size () < 2 || ! (c_line -> FromString (tokens[1])) ) {
	manager -> ParserWarning ("missing or malformed channel");
	c_line -> Set ();
      } else if (tokens.size () > 2)
	manager -> ParserWarning ("ignoring trailing arguments.");

      current -> childs.push_back (c_line);

      return true;

    } else if (tokens[0]=="essid") {
      current = GetWirelessSubsection (manager);

      EssidLine* e_line = new EssidLine (content);
      if (tokens.size () < 2 ) {
	manager -> ParserWarning ("missing id, assuming 'any'");
	e_line -> Set ();
      } else {
	if (tokens.size () > 2)
	  manager -> ParserWarning ("ignoring trailing arguments.");
	e_line -> essid = (tokens[1] == "any") ? "" : tokens[1];
	e_line -> Set ();
      }

      current -> childs.push_back (e_line);

      return true;

    } else if (tokens[0]=="key") {
      current = GetWirelessSubsection (manager);

      KeyLine* k_line = new KeyLine (content);
      if (tokens.size () < 2 ) {
	manager -> ParserWarning ("missing key, assuming 'off'");
	k_line -> Set ();
      } else {
	if (tokens.size () > 2)
	  manager -> ParserWarning ("ignoring trailing arguments.");
	k_line -> key = (tokens[1] == "off") ? "" : tokens[1];
	k_line -> Set ();
      }

      current -> childs.push_back (k_line);

      return true;
    }
    
    return false;
  }




  PPPSubsection* PPPSubsectionInstance::GetPPPSubsection (kvtree::TreeManager* manager)
  {
    if (!EnsureInterfaceSubsection (manager, kvtree::node_types::ppp_subsection) )
      manager -> OpenNewSubsection (new PPPSubsection (manager -> getCurrent ()));
    
    kvtree::Group* ret = manager -> getCurrent ();
    assert (ret -> Type () == kvtree::node_types::ppp_subsection);
    return (PPPSubsection*) ret;
  }



  bool PPPSubsectionInstance::Match (std::string& content, kvtree::TreeManager* manager)
  {
    PPPSubsection* current;

    if (tokens[0]=="ppp") {
      current = GetPPPSubsection (manager);

      DeviceLine* d_line = new DeviceLine (content);
      if (tokens.size () < 2 ) {
	manager -> ParserWarning ("missing device, assuming '/dev/tty'");
	d_line -> device = "/dev/tty";
	d_line -> Set ();
      } else {
	if (tokens.size () > 3)
	  manager -> ParserWarning ("ignoring trailing arguments.");
	if (tokens.size () > 2 && ! d_line -> SpeedFromString (tokens[2]))
	  manager -> ParserWarning ("invalid speed, setting to default (none)");

	d_line -> device = tokens[1];
      }
      
      current -> childs.push_back (d_line);

      return true;

    } else if (tokens[0]=="pppoe") {
      current = GetPPPSubsection (manager);
      current -> SetOverEthernet (true);
      return true;
 
    } else if (tokens[0]=="ppp_defaults") {
      current = GetPPPSubsection (manager);
      current -> SetDefaults (true);
      return true;
 
    } else if (tokens[0]=="chat_defaults") {
      current = GetPPPSubsection (manager);
      current -> SetChatDefaults (true);
      return true;
 
    } else if (tokens[0]=="ppp_speed_defaults") {
      current = GetPPPSubsection (manager);
      current -> SetSpeedDefaults (true);      
      return true;
 
    } else if (tokens[0]=="ppp_option") {
      current = GetPPPSubsection (manager);

      if (tokens.size() < 2)
	manager -> ParserWarning ("ignoring empty option");
      else {
	OptionLine* o_line = new OptionLine (content);
	o_line -> option = tokens[1];
	if (tokens.size () > 2)
	  o_line -> value = tokens[2];

	o_line -> Set ();

	if (tokens.size () > 3)
	  manager -> ParserWarning ("ignoring trailing arguments.");
	
	current -> childs.push_back (o_line);
      }

      return true;
 
    } else if (tokens[0]=="chat_init") {
      current = GetPPPSubsection (manager);

      if (tokens.size() < 2)
	manager -> ParserWarning ("ignoring empty init string");
      else {
	ChatInitLine* ci_line = new ChatInitLine (content);
	ci_line -> init = tokens[1];
	ci_line -> Set ();

	if (tokens.size () > 2)
	  manager -> ParserWarning ("ignoring trailing arguments.");
	
	current -> childs.push_back (ci_line);
      }

      return true;
 
    } else if (tokens[0]=="chat_dial") {
      current = GetPPPSubsection (manager);

      if (tokens.size() < 2)
	manager -> ParserWarning ("ignoring missing telephone number");
      else {
	ChatDialLine* cd_line = new ChatDialLine (content);
	cd_line -> dial = tokens[1];
	cd_line -> Set ();

	if (tokens.size () > 2)
	  manager -> ParserWarning ("ignoring trailing arguments.");
	
	current -> childs.push_back (cd_line);
      }

      return true;
 
    }

    return false;
  }





  bool SortRemainingInstance::Match (std::string& content, kvtree::TreeManager* manager)
  {
    if (manager -> getCurrent () -> Type () != kvtree::node_types::global)
      EnsureInterfaceSubsection (manager, kvtree::node_types::interface_section);
    return false;
  }

}

